/***********************************************************************

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

***********************************************************************/
package gecco.game.dke;

/**
 * MarchAction.java
 *
 *
 * Created: Thu Oct 17 13:16:31 2002
 *
 * @author <a href="mailto:hans.sandstrom@fhs.mil.se">Hans Sandstrm</a>
 * @version
 */

import gecco.server.core.*;
import gecco.server.automaton.Automaton;
import gecco.server.visibility.Point;
import gecco.server.core.Server;
import java.util.Vector;

public class MarchAction extends MoveAction {

    Point[] marchRoute;
    int marchRouteIndex = 0;
    Point start;

    public MarchAction (){
	super();
    }
    
    public void initiate(int _actionHandle, int _unitHandle) {
	throw new IllegalArgumentException("This action must be instanciated with a point as an argument.");
    }

    /**
     *  Initialise the action with a point as argument by calling
     *  super.initiate(_actionHandle, _unitHandle, _argX, _argY).
     *  Then check if there is a route to the endpoint.
     *
     *
     * @param  _actionHandle  A unique handle that should identify this action.
     * @param  _unitHandle    The handle of the executing unit.
     * @param  _argX          The x coordinate of the target point.
     * @param  _argY          The y coordinate of the target point.
     */
    public void initiate(int _actionHandle, int _unitHandle, double _argX, double _argY) {
	super.initiate(_actionHandle, _unitHandle, _argX, _argY);
	if (message != null) {
	    return;
	}

	// Check if start pos is on a route
	int startIntX = ReferenceHolder.getUnitManager().getUnit(_unitHandle).getIntX();
	int startIntY = ReferenceHolder.getUnitManager().getUnit(_unitHandle).getIntY();
	Automaton a = ReferenceHolder.getAutomatonMap().getAutomaton(startIntX, startIntY);
	AutomatonImpl startAutomaton = (AutomatonImpl) a;
	if (!startAutomaton.isRoute()) {
	    message = Def.getMustStartOnRouteString();
	    return;
	}

	// Check if end pos is on a route
	int endIntX = (int) _argX;
	int endIntY = (int) _argY;
	a = ReferenceHolder.getAutomatonMap().getAutomaton(endIntX, endIntY);
	AutomatonImpl endAutomaton = (AutomatonImpl) a;
	if (!endAutomaton.isRoute()) {
	    boolean automatonIsRoute = false;
	    for (int dy  = -1; dy < 2 && !automatonIsRoute; dy++) {
		for (int dx = -1; dx < 2 && !automatonIsRoute; dx++) {
		    a = ReferenceHolder.getAutomatonMap().getAutomaton(endIntX+dx, endIntY+dy);
		    endAutomaton = (AutomatonImpl) a;
		    if (endAutomaton.isRoute()) {
			setArgumentX(_argX+dx);
			setArgumentY(_argY+dy);
			automatonIsRoute = true;
		    }
		}
	    }
	    if (!automatonIsRoute) {
		message = Def.getMustEndOnRouteString();
		return;
	    }
	}

	marchRoute = RouteFinder.findRoute(startAutomaton, new Point(startIntX, startIntY),
					   endAutomaton, new Point(endIntX, endIntY));
	if (marchRoute == null) {
	    message = "No route found!";
	    return;
	}
	start = new Point(startIntX, startIntY);
    }

    public void initiate(int _actionHandle, int _unitHandle, int _argUnitHandle) {
	throw new IllegalArgumentException("This action must be instanciated with a point as an argument.");
    }

    /**
     *  This method is called periodically by the game engine.
     *
     * @return    An object of type ActionReturn, which encapsulates data about
     *      the status of the action after the checkpoint. Different instances
     *      of the class ActionReturn should be returned depending on whether
     *      the action is completed, aborted, or if the action should be
     *      requeued and the checkpoint method called again at some time later.
     */
    public ActionReturn checkPoint(){
	if (getWarUnit().isInCombat() || getWarUnit().underAttackInitiated 
	    || getWarUnit().attackInitiated || getWarUnit().isFiringIndirect()) {
	    return new ActionReturnRequeue(5.0);
	}
	
	if (firstRun) {
	    if (message != null) {
		return new ActionReturnError(message);
	    }
	    int x = getWarUnit().getIntX();
	    int y = getWarUnit().getIntY();
	    markAsVisited(x, y);
	    expandVisitedToRouteWidth(x,y,(Math.abs(addX) > Math.abs(addY)? 1:0));
	    
	    updateUnitStatus(true);
	    firstRun = false;
	}
	
	if (marchRouteIndex >= marchRoute.length) {
	    if (marchRouteIndex != 0) {
		// Debug info
		System.out.println("MarchAction: Ngot r konstigt, marchRouteIndex = " + marchRouteIndex + 
				   "marchRoute.length = " + marchRoute.length);
	    }
	    return new ActionReturnError();
	}
	//Goto center of next pixel on route
	double nextX = marchRoute[marchRouteIndex].x + 0.5;
	double nextY = marchRoute[marchRouteIndex].y + 0.5;
	
	if (isCloseEnough(nextX, nextY)) {
	    if (!(getWarUnit().getIntX() == (int)nextX &&
		  getWarUnit().getIntY() == (int)nextY)) {
		markAsVisited((int)nextX, (int)nextY);
	    }

	    marchRouteIndex++;

	    // Destination reached
	    if (marchRouteIndex >= marchRoute.length) {
		updateUpdateVisited();
		// getWarUnit().setPosition(nextX, nextY);
		getWarUnit().setPosition(nextX, nextY, true);
		testUpdateVisited();
		updateUnitStatus(false);
		return new ActionReturnCompleted(Def.getDestinationReachedString());
	    }
	    
	    nextX = marchRoute[marchRouteIndex].x + 0.5;
	    nextY = marchRoute[marchRouteIndex].y + 0.5;
	}
	
	double curX = getWarUnit().getX();
	double curY = getWarUnit().getY();

	// Add checks for curX+addX and curY+addY before the move are executed 

	AutomatonImpl currAutomaton = (AutomatonImpl)ReferenceHolder.getAutomatonMap().
	    getAutomaton(getWarUnit().getIntX(), getWarUnit().getIntY());

	double stepLength;

	if (currAutomaton.isRoute()) {
	    double dx = nextX - curX;
	    double dy = nextY - curY;
	    stepLength = Math.sqrt(dx*dx + dy*dy);
	    if (stepLength > 2.2) {
		stepLength = Def.STEP_LENGTH;
	    }
	}
	else {
	    stepLength = Def.STEP_LENGTH;
	}
	calcAddXAddY(nextX, nextY, stepLength);

	if (getWarUnit().cantMoveTo(getAutomatonState((int)(curX+addX), (int)(curY+addY)))) {
	    updateUnitStatus(false);
	    return new ActionReturnError(Def.getCantMoveToWaterString());
	}
	
	updateUpdateVisited();

	// Step forward
	curX = curX + addX;
	curY = curY + addY;
	getWarUnit().setPosition(curX, curY, true);

	testUpdateVisited();

	return new ActionReturnRequeue(getWarUnit().getTimeToRequeueForMovement(stepLength));
    }

    public void updateUnitStatus(boolean b){
	getWarUnit().setMarching(b);
	getWarUnit().updateStatus();
    }

    public void prepareAbortion(){
	updateUnitStatus(false);
    }

}// MarchAction
